/*
 * Copyright (C) 2000-2011 OSS Nokalva, Inc.  All rights reserved.
 *
 * THIS FILE IS PROPRIETARY MATERIAL OF OSS NOKALVA AND MAY BE USED ONLY
 * BY DIRECT LICENSEES OF OSS NOKALVA. THIS FILE MAY NOT BE DISTRIBUTED.
 */

/*
 * This program illustrates how the encoding/decoding of a huge SEQUENCE OF
 * field of a PDU can be achieved using files instead of keeping the entire
 * unencoded/decoded PDU in memory.
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "iaapi.h"
#include "sample.h"

int startEncodeSeqOf2File(OssGlobal *, int, FILE *, int *, int *);
int endEncodeSeqOf2File(OssGlobal *, FILE *, long, int, int);
int decodeSeqOfFromFile(OssGlobal *, int, FILE *, long *, long *);


/*
 * Initialize statically our outer enclosing PDU.  Not all the field values yet.
 */
static char _v0[] = "IA5String";
static struct OuterPDU_sof_encodable *_v2 = NULL;
OuterPDU outerPDU = {
    _v0,
    {OuterPDU_sof_encodable_PDU, 0, NULL, &_v2}
};


#define FILE_NAME_LENGTH 128

/*
 * Temporary file for holding the entire deferred SEQUENCE OF encoded field.
 */
#define SEQOF_FILE_NAME "sof.ber"

/*
 * File for holding the outermost entire encoded PDU value.
 */
#define OUTER_PDU_FILE_NAME "outer.ber"

/*
 * Number of elements in the SEQUENCE OF field of the outer enclosing PDU.
 */
#define NUMBER_OF_ELEMENTS 10



int main(void)
{

    OssGlobal w,                /* OSS's global structure */
	     *world = &w;       /* pointer to OSS's global structure */
    int       outer_pdu_num;    /* outer PDU number */
    int       i, rc;            /* return code */
    char      outerPduFileName[FILE_NAME_LENGTH];
    char      sofFileName[FILE_NAME_LENGTH];
    OuterPDU *decodedPDU = NULL;
    OssBuf    encodedPDU;       /* length and address of encoded data */
    int       sof_pdu_num = OuterPDU_sof_encodable_PDU;
    int       element_pdu_num = I_PDU;
    OpenTypeExtended *sof = &outerPDU.sof;
    FILE     *fp;
    long      sof_offset, sof_length;
    int       indef;
    int       eoc = 0;
    void     *decodedSOF = NULL;


	/*
	 * Call ossinit() first to initialize OSS's resources.
	 */
    if ((rc = ossinit(world, sample))) {
	printf("Ossinit() returned %d.\n", rc);
	return 1;
    }

	/*
	 * Set the encoding rules and flags.  It is enough to set DEFINITE or
	 * INDEFINITE here to switch between the 2 encoding types.
	 */
    ossSetEncodingRules(world, OSS_BER);
    ossSetEncodingFlags(world, DEBUGPDU+DEFINITE);
    ossSetDecodingFlags(world, DEBUGPDU);
    indef = INDEFINITE & ossGetEncodingFlags(world);

	/*
	 * Preparing for encoding and decoding.
	 */
    outer_pdu_num = OuterPDU_PDU;	/* Outermost PDU identifier */
    strncpy(outerPduFileName, OUTER_PDU_FILE_NAME, sizeof(outerPduFileName));
    strncpy(sofFileName, SEQOF_FILE_NAME, sizeof(sofFileName));

	/*
	 * Encode the deferred SEQUEBNCE OF field to a file...
	 */
    ossPrint(world, "Encoding the deferred SEQUENCE OF field to a file...\n\n");
    fp = fopen(sofFileName, "wb");
    if (startEncodeSeqOf2File(world, sof_pdu_num, fp, (int *)&sof->length, &eoc))
	return 1;
    fclose(fp);
	/*
	 * Mark the "encoded" field as a file object to tell the encoder
	 * to encode the SEQUENCE OF field of the outer PDU from the file
	 * sof.ber.
	 */
    sof->encoded = (unsigned char *)ossMarkObj(world, OSS_FILE, sofFileName);
    sof->pduNum = sof_pdu_num;
    sof->decoded = NULL;
    sof->byteOffset = 0;
    sof->bitOffset = 0;
    sof->residualBits = 0;

    rc = ossOpenEncoderOutputFile(world, sofFileName);

	/*
	 * The sof->length is the number of octets written to the file so
	 * it is actually an offset to the file indicating where to continue
	 * writing encoded data.
	 */
    encodedPDU.length = sof->length;
    encodedPDU.value = ossMarkObj(world, OSS_FILE, sofFileName);

	/*
	 * Encode all the SEQUENCE OF elements one by one to a temporary
	 * file.  The encoder updates the offset encodedPDU->length to the
	 * end of the file.  So the loop is simple.
	 */
    for (i = 0; i < NUMBER_OF_ELEMENTS; i++) {
	I element = i; /* this is INTEGER in our case for simplicity */

	/*
	 * Encoding the element of the SEQUENCE OF field of the outer
	 * enclosing PDU.
	 */
	if ((rc = ossEncode(world, I_PDU, &element, &encodedPDU))) {
	    ossPrint(world, "Encoding failed with error = %d.\n", rc);
	    return 1;
	}
    }
    sof_length = encodedPDU.length - sof->length;
    ossCloseEncoderOutputFile(world);

	/*
	 * Finish encoding the inner SEQUENCE OF value to the temporary file.
	 */
    fp = fopen(sofFileName, "r+b");
    rc = endEncodeSeqOf2File(world, fp, sof_length, sof->length, eoc);
    fclose(fp);

	/*
	 * Set the total length of the encoded SEQUENCE OF field of the
	 * outer enclosing PDU by adding the number of octets occupied by
	 * the eoc zero octets for INDEFINITE-length encoding.
	 */
    sof->length += sof_length + eoc * 2;

    ossPrint(world, "\n\tPrinting the unencoded outermost enclosing PDU...\n\n");
    ossPrintPDU(world, outer_pdu_num, &outerPDU);

	/*
	 * Encode the outermost enclosing PDU to a file, outer.ber.
	 */
    encodedPDU.length = 0;
    encodedPDU.value = (unsigned char *)ossMarkObj(world, OSS_FILE, outerPduFileName);
    ossPrint(world, "\nEncoding the outer enclosing PDU to a file...\n\n");
    if ((rc = ossEncode(world, outer_pdu_num, &outerPDU, &encodedPDU))) {
	ossPrint(world, "Encoding failed with error = %d.\n", rc);
	return 1;
    }

	/*
	 * Decode the outermost enclosing PDU from the file outer.ber.
	 */
    encodedPDU.length = 0;
    encodedPDU.value = (unsigned char *)ossMarkObj(world, OSS_FILE, outerPduFileName);
    ossPrint(world, "\nDecoding the outer enclosing PDU from the file...\n\n");
    if ((rc = ossDecode(world, &outer_pdu_num, &encodedPDU, (void **)&decodedPDU))) {
	ossPrint(world, "Decoding failed with error = %d.\n", rc);
	return 1;
    }

    ossPrint(world, "\n\tPrinting the decoded outermost enclosing PDU...\n\n");
    ossPrintPDU(world, outer_pdu_num, decodedPDU);

	/*
	 * Decode the deferred SEQUENCE OF field from the file.
	 */
    ossPrint(world, "\nDecoding the deferred SEQUENCE OF field from the file...\n\n");

    if (!(fp = fopen(outerPduFileName, "rb"))) {
	ossPrint(world, "File '%s' failed to open.\n", outerPduFileName);
	return 1;
    }

	/*
	 * Set the outer PDU file position to the beginning of the SEQUENCE OF
	 * field of the outer enclosing PDU.  We know the file offset now
	 * because the decoder returned the offset in the
	 * decodedPDU->sof.byteOffset.
	 */
    if (fseek(fp, decodedPDU->sof.byteOffset, SEEK_SET)) {
	ossPrint(world, "Cannot set the outer PDU file position.\n");
	fclose(fp);
	return 1;
    }
    if (decodeSeqOfFromFile(world, sof_pdu_num, fp, &sof_offset, &sof_length))
	return 1;
    fclose(fp);

	/*
	 * Get the start offset for the contents of the SEQUENCE OF field
	 * of the outermost enclosing PDU.  In our case it is concatenated
	 * INTEGER PDUs.  But it can be encoded PDUs of any other ASN.1 type.
	 */
    sof_offset += decodedPDU->sof.byteOffset;

    ossOpenDecoderInputFile(world, outerPduFileName);

	/*
	 * Tell the decoder to start decoding from the file object sof.ber
	 * at an offset of "sof_offset".
	 */
    encodedPDU.value = decodedPDU->sof.encoded;
    encodedPDU.length = sof_offset;

	/*
	 * Decode all the elements one by one of the inner SEQUENCE OF
	 * field, INTEGER values in our case.  In case of INDEFINITE-length
	 * encoding, you do not know the exact total length occupied by
	 * the encoded elements of the SEQUENCE OF field.  So an attempt to
	 * decode past the last element of the SEQUENCE OF results in the
	 * following decoder returning error code 5 with error messages:
	 *
	 * D0058E: Expected tag, found EOC octets.
	 * D0034E: PDU and input tags do not match: [UNIVERSAL 0]; check PDU #1.
	 */
    while (sof_length > 0 || indef) {
	I *element;
	
	element = NULL;

	rc = ossDecode(world, &element_pdu_num, &encodedPDU, (void **)&element);
	if (rc)
		 /*
		  * In case of INDEFINITE-length encoding, break
		  * because no more encoded PDUs are left in the file.
		  */
	    break;
	i = ossGetBytesReadByDecoder(world);
	if (!indef)
	    sof_length -= i;
	encodedPDU.length += i;
	ossFreePDU(world, element_pdu_num, element);
    }
    ossCloseDecoderInputFile(world);

    ossFreePDU(world, outer_pdu_num, decodedPDU);

	/*
	 * Now to be 100% sure that we encoded the SEQUENCE OF field of the
	 * outer enclosing PDU correctly, call the decoder normally to decode
	 * the SEQUENCE OF PDU from the file sof.ber to memory in one step.
	 */
    encodedPDU.length = 0;
    encodedPDU.value = (unsigned char *)ossMarkObj(world, OSS_FILE, sofFileName);
    ossPrint(world, "\nDecoding the SEQUENCE OF value from the file as a standalone PDU...\n\n");
    rc = ossDecode(world, &sof_pdu_num, &encodedPDU, &decodedSOF);
    if (rc) {
	ossPrint(world, "Decoding failed with error = %d.\n", rc);
	return 1;
    }
    ossFreePDU(world, sof_pdu_num, decodedSOF);

	/*
	 * Free the resources.
	 */
    ossterm(world);
    remove(sofFileName);	/* remove the temporary SEQUENCE OF file */
    remove(outerPduFileName);	/* remove the file with the outermost PDU */
    return 0;
}


/*
 * Maximum length of the encoded tag & length pair.
 */
#define MAX_TAG_LENGTH_PAIR (((sizeof(long) * 8) / 7) + 3 + sizeof(long))
/*
 * Function:            encodes a constructed tag number, not class and
 *                      constructedness.
 *
 * Parameters:
 *
 *   "start"        ... pointer to buffer with encoded tag
 *   "tag"          ... tag value to encode
 *
 * Return value:        number of octets of encoded tag
 */
static size_t encodeMultiBytesTagNumber(unsigned char *start, unsigned int tag)
{

    size_t	len = 1;
    unsigned char *end;


    *start++ = 0x1F;	/* 0001 1111  - value of the first byte */
    end = start;

	/*
	 * Now fill the array with 7-bit fragments of '1ppp pppp'.
	 * The array will be filled lower fragment first.
	 */
    do {
	*end++ = 0x80 | (tag & 0x7F);
	len++;		/* count length */
    } while (tag >>= 7);
    *start &= 0x7F;	/* least significat octet should have zero highest bit */

	/*
	 * Swap buffer to make lower octet (with fragment of tag) last.
	 */
    while (start < end) {
	unsigned char tmp = *start;

	/*
	 * Here "start" points to the first octet, "end" to the next octet
	 * after the last one.
	 */
	*start++ = *--end;
	*end = tmp;
    }
    return len;
}


/*
 * Function:            encodes a tag of a PDU with the "pdu_num" into an OssBuf
 *                      buffer.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "pdu_num"      ... PDU number
 *   "buffer"       ... poiner to OSS's OssBuf buffer
 *   "index"        ... tag index
 *   "th"           ... type handle
 *
 * Return value:        zero if successful and 1 on error
 */
static int encodeTagOfType(OssGlobal *world,
		int pdu_num, OssBuf *buffer, int index, TypeHndl th)
{

    int            tag;
    ASN1TagClass   tag_class;
    unsigned char *bytes = buffer->value;


    if (!(tag = ossAsn1TagNumber(world, th, index))) {
	ossPrint(world, "encodeTagOfType(): incorrect field type.\n");
	buffer->length = 0;
	return 1;
    }

    if ((tag_class = ossAsn1TagClass(world, th, index)) == NULLENCODING) {
		/*
		 * In this case we should not encode the tag at all.
		 */
	buffer->length = 0;
	return 1;
    }

	/*
	 * Here a tag is a tag number without class and constructed bits
	 * A tag class is a class of the tag (UNIVERSAL, PRIVATE,... etc.)
	 */
	 
	 /*
	  * Encode the tag number itself.
	  */
    if (tag < 31) {		/* may fit it 5 bits, excluding '11111' */
	bytes[0] = tag;		/* can be encoded in one byte */
	buffer->length = 1;
    }
    else
	buffer->length = encodeMultiBytesTagNumber(bytes, (unsigned int)tag);
    
	/*
	 * Set the constructed bit.
	 */
    bytes[0] |= 0x20;

	/*
	 * Then set the tag class.
	 */
    switch (tag_class) {
	case UNIVERSAL:
	    bytes[0] |= 0x00;
	    break;
	case APPLICATION:
	    bytes[0] |= 0x40;
	    break;
	case CONTEXT_SPECIFIC:
	    bytes[0] |= 0x80;
	    break;
	case PRIVATE:
	    bytes[0] |= 0xC0;
    }

    return 0;
}


/*
 * Function:            encodes a length using the long form encoding:
 *                      1Nnn nnnn  LLLL LLLL  LLLL LLLL  ....  LLLL LLLL
 *                       hi    lo      1st        2nd    ....      Nth
 *                      where 'Nnn nnnn' is a positive number of subsequent
 *                      octets representing the value of the length field.
 *                      Leftmost bit of the leftmost first octet is the most
 *                      significant.  The encoded length is stored in the
 *                      user's provided buffer pointed by 'buffer'.
 *
 * Parameters:
 *
 *   "length"       ... length value to be represented in the long for of encoding
 *   "buffer"       ... pointer to buffer to store the result
 *
 * Return value:        number of bytes written to the buffer
 */
static int encodeDefiniteLength(unsigned long length, unsigned char *buffer)
{

    int i;

	/*
	 * To store the 'long' length sizeof(length) + 1 bytes is needed.
	 */
    for (i = sizeof(length) ; i > 0; i--) {
	buffer[i] = (unsigned char)(length & 0x000000FF);
	length >>= 8;
    }

	/*
	 * The very first octet is 1nnn nnnn, where 7 lower bits is the
	 * number of the subsequent octets.
	 */
    *buffer = 0x80 | sizeof(length);
    
    return sizeof(length) + 1;
}


/*
 * Function:            starts encoding of the SEQUENCE OF field of the outer
 *                      enclosing PDU.  Writes all the tag=length pairs to the
 *                      open temporary file.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "pdu_num"      ... PDU number of the SEQUENCE OF type
 *   "fp"           ... file pointer
 *   "length"       ... total length value written to be returned
 *   "eoc"          ... number of eoc's to be returned for INDEFINITE-length
 *                      encodings
 *
 * Return value:        zero if successful and 1 on error
 */
int startEncodeSeqOf2File(OssGlobal *world,
			int pdu_num, FILE *fp, int *length, int *eoc)
{

    OssBuf    buffer;
    int       rc, number_of_tags, i;
    void     *saved;
    TypeHndl  th;


	/*
	 * Get the type handle for the SEQUENCE OF type of the
	 * outer enclosing PDU.
	 */
    th = ossTypeHandleOfPDU(world, pdu_num);

    if (!(number_of_tags = ossNumberOfTags(world, th))) {
	ossPrint(world, "Zero number of tags was returned by ossNumberOfTags().\n");
	buffer.length = 0;
	return 1;
    }

    buffer.length = 0;
    buffer.value = saved = malloc(MAX_TAG_LENGTH_PAIR * number_of_tags);

    for (i = 0; i < number_of_tags; i++) {

	if (rc = encodeTagOfType(world, pdu_num, &buffer, i, th))
	    return rc;

	/*
	 * Write length for DEFINITE- or 0x80 for INDEFINITE-length
	 * encodings.
	 */
	if (INDEFINITE & ossGetEncodingFlags(world)) {
	    buffer.value[buffer.length++] = 0x80;
	    (*eoc)++;
	}
	else
	    buffer.length += encodeDefiniteLength(0, buffer.value + buffer.length);

		/*
		 * Write the encoded tag & length pair.
		 */
	if (fwrite(buffer.value, 1, buffer.length, fp) != (size_t)buffer.length) {
	    ossPrint(world, "Error writing to the temporary SEQUENCE OF file.\n");
	    return 1;
	}
	*length += buffer.length;
	buffer.value += buffer.length;
    }
    free(saved);
    return 0;
}


/*
 * Function:            ends encoding of the SEQUENCE OF field of the outer
 *                      enclosing PDU.  Writes all the eoc's for INDEFINITE-
 *                      and tag-length pairs for DEFINITE-length encodings to
 *                      the open temporary file.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "fp"           ... file pointer
 *   "to*_el*_l*th" ... total length og all the encoded elements of the
 *                      SEQUENCE OF of the outer enclosing PDU
 *   "le*_written*" ... length of the tag-length pairs of the SEQUENCE OF
 *                      field of the outer enclosing PDU
 *   "eoc"          ... number of eoc's to be returned for INDEFINITE-length
 *                      encodings
 *
 * Return value:        zero if successful and 1 on error
 */
int endEncodeSeqOf2File(OssGlobal *world, FILE *fp,
			long total_element_length, int length_written, int eoc)
{

    int            count, i, rc = 0;
    unsigned char *head, *byte;
    long           encLengthSize;
    long           length_left = length_written;
    

    if (INDEFINITE & ossGetEncodingFlags(world)) {
		/*
		 * For INDEFINITE-length the octet 0x80 is written before.
		 * So now we need to match each 0x80 with 2 0x00's.
		 */
	if (fseek(fp, 0, SEEK_END)) {
	    ossPrint(world, "Cannot set the SEQUENCE OF temporary file position.\n");
	    return 1;
	}
	for (i = 0; i < 2 * eoc; i++) {
	    if (fputc(0, fp) == EOF) {
		ossPrint(world, "Error writing to the temporary SEQUENCE OF file.\n");
		rc = 1;
		break;
	    }
	}
    }
    else {
	head = byte = malloc(length_left);
	if (fread(byte, 1, length_left, fp) != (size_t)length_left) {
	    ossPrint(world, "Error reading from the temporary SEQUENCE OF file.\n");
	    rc = 1;
	}

	while (length_left > 0) {
	    count = 1;
		/*
		 * If five lower bits are all 1's the encoding of tag is long.
		 * Skipping all subsequent octets with set highest bit, and
		 * one with reset (it is the last octet of the tag). The
		 * 'count' == 1 at the start (length of initial octet of the
		 ( tag).
		 */
	    if ((*byte & 0x1F) == 0x1F) { /* xxx1 1111 */
		do {
		    byte++;
		    count++;		/* count this octet */
		} while (*byte & 0x80);	/* this one was the last octet of the tag */
	    }
	    byte++;
	    length_left -= count;

		/*
		 * For DEFINITE-length length no eoc's are written.  But we
		 * need to write the correctly calculated lengths to their
		 * places.  We go from front to end because the length of
		 * the length field is constant and always 5 in our case.
		 */
	    encLengthSize = encodeDefiniteLength(length_left +
						total_element_length - 5, byte);
	    length_left -= encLengthSize;
	    byte += encLengthSize;
	}
	if (fseek(fp, 0, SEEK_SET)) {
	    ossPrint(world, "(): Cannot set the Length position in the input file.\n");
	    rc = 1;
	}
	if (fwrite(head, 1, length_written, fp) != (size_t)length_written) {
	    ossPrint(world, "Error writing to the temporary SEQUENCE OF file.\n");
	    rc = 1;
	}
        free(head);
    }
    return rc;
}


/*
 * Function:            reads a tag-length pair from a file and returns the
 *                      number of octets read for DEFINITE- and -1 for
 *                      INDEFINITE-length encodings.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "fp"           ... file pointer
 *   "length"       ... decoded length to be returned
 *
 * Return value:        non-zero length of pair if successful and 0 on error
 */
static size_t getPairLengthFromFile(OssGlobal *world, FILE *fp, long *length)
{

    size_t	skip;	/* size of encoded tag & length to skip */
    int		byte;


	/*
	 * Read first octet of encoded tag (issue an error if failed).
	 */
    if ((byte = fgetc(fp)) == EOF) {
	ossPrint(world, "Error reading file.\n");
	return 0;
    }
    skip = 1;

	/*
	 * If five lower bits are all 1's, the encoding of tag is a long form.
	 * Skipping all subsequent octets with highest bits set.
	 */
    if ((byte & 0x1F) == 0x1F) {
	do {
	    if ((byte = fgetc(fp)) == EOF) {
		ossPrint(world, "Error reading file.\n");
		return 0;
	    }
	    skip++;		/* count this octet */
	} while (byte & 0x80);	/* this one was the last octet of the tag */
    }
    
    if ((byte = fgetc(fp)) == EOF) {
	ossPrint(world, "Error reading file.\n");
	return 0;
    }

    skip++;		/* skip this octet */
    if (byte == 0x80)   /* indicates INDEFINITE length */
	*length = -1;	/* means INDEFINITE */
    else {
	if (byte & 0x80) { /* higher bit set indicates long form */
	    int len_of_len = byte & 0x7F;
	    
	    skip += len_of_len;	/* number bytes used for subsequent octets of "length" */
	    *length = 0;
	    while (len_of_len--) {
		if ((byte = fgetc(fp)) == EOF) {
		    ossPrint(world, "Error reading file.\n");
		    return 0;
		}
		skip++;
		*length = ((unsigned long)*length << 8) + byte;
	    }
	}
	else
	    *length = byte;
    }
    return skip;
}


/*
 * Function:            reads a tag-length pair from a file and returns the
 *                      number of octets read for DEFINITE- and -1 for
 *                      INDEFINITE-length encodings.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "pdu_num"      ... PDU number of the SEQUENCE OF type
 *   "fp"           ... file pointer
 *   "sof_offset"   ... offset--to be returned--to the beginning of the encoded
 *                      elements of the SEQUENCE OF field of the outer
 *                      enclosing PDU
 *   "sof_length"   ... total length--to be returned--of the encoded elements
 *                      of the SEQUENCE OF field of the outer enclosing PDU
 *
 * Return value:        zero if successful and 1 on error
 */
int decodeSeqOfFromFile(OssGlobal *world, int pdu_num, FILE *fp,
				long *sof_offset, long *sof_length)
{

    size_t    tagAndLengthSize;
    long      length_decoded;
    TypeHndl  th;
    int       i, numberOfTags;


    *sof_offset = *sof_length = 0;
    th = ossTypeHandleOfPDU(world, pdu_num);

    if (!(numberOfTags = ossNumberOfTags(world, th))) {
	ossPrint(world, "Zero number of tags was returned by ossNumberOfTags().\n");
	return 1;
    }
    for (i = 0; i < numberOfTags; i++) {
	/*
	 * Get the length of the encoded tag-length pair.
	 */
	if (!(tagAndLengthSize = getPairLengthFromFile(world, fp, &length_decoded)))
	    return 1;
	*sof_offset += tagAndLengthSize;
    }
    *sof_length = length_decoded;
    return 0;
}

